#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

# TODO: Typing hints

from typing import TypeVar, Tuple
import warnings

import casadi as ca
import numpy as np

from .util import check_and_wrap_to_list, check_if_has_duplicates, check_if_list_of_type, check_if_list_of_none, \
    check_if_square, who_am_i

Symbolic = TypeVar('Symbolic', ca.SX, ca.MX)


class GenericCost:
    """Class for generic cost functions"""

    def __init__(self, model, use_sx=True):
        """Constructor method"""
        self._cost = 0
        self._is_set = False
        self._model = model
        self._use_sx = use_sx

        # Number of time-varying references. Use for traj. tracking and path following
        self._n_tv_ref = 0

    def __repr__(self):
        """Representation method"""
        pass

    @property
    def cost(self):
        """

        :return:
        """
        return self._cost

    @cost.setter
    def cost(self, arg):
        # TODO check if the arg has the same SX variables that appear in the model
        if not isinstance(arg, ca.SX) and not isinstance(arg, ca.MX):
            raise TypeError('The cost function must be a casadi SX/MX expression.')

        if not arg.shape[0] == arg.shape[0] == 1:
            raise TypeError('The cost must be a one-dimensional function.')

        # Substitute the measurement equation if the user imputed it
        for i in range(self._model.n_y):
            arg = ca.substitute(arg, self._model._y[i], self._model.meas[i])

        self._is_set = True
        self._cost += arg

    @property
    def n_tv_ref(self):
        """
        Number of time-varying references

        :return:
        """
        return self._n_tv_ref


class QuadraticCost(GenericCost):
    """"""

    def __init__(self, model, use_sx=True):
        """Constructor method"""
        super().__init__(model, use_sx=use_sx)

        self._paths_list = []
        self._trajectories_list = []
        self._references_list = []
        self._input_change_list = []

        self._has_trajectory_following = False
        self._has_path_following = False
        # Number of time-varying references. Use for traj. tracking and path following
        self._n_tv_ref = 0

        # Placeholders for the references, needed for trajectory-tracking problems
        self.ref_placeholder = ca.SX.sym('foo', 0)

        # Placeholders for the inputs changes
        self.input_change_placeholder = []

        # Vector that stores the indices of the inputs that enter the change input cost term
        self._ind_input_changes = []

        # Added variable - used for checking
        # TODO: use them in the different add things
        self._added_states = []

        # Store the name of all the varying trajectories
        self._varying_trajectories_names = []

        # Store the name of all the varying trajectories for which the trajectory has not yet been provided
        self._open_varying_trajectories_names = []

        # Initialize weighting matrices
        self._Q = None
        self._R = None
        self._P = None

    @staticmethod
    def _check_dimensions(var_list, var_weight, var_ref, type):
        """
        Run some sanity checks on the user supplied values

        :param var_list:
        :param var_weight:
        :param var_ref:
        :param type:
        :return:
        """
        if var_list is not None and var_weight is None:
            raise ValueError(f"You passed the following {type}: {var_list} to the cost function, but I do not have any "
                             f"weights for it/them. Please pass me the weights.")

        if var_list is None and var_weight is not None:
            raise ValueError(f"You the passed the {type} weights {var_weight} but without states names. Please pass me "
                             f"the states names with quad_stage_cost.x.")

        if not check_if_list_of_none([var_list, var_weight]):
            if not len(var_list) == var_weight.shape[0]:
                raise ValueError(f"{type} and weights dimensions must be compatible. The {type} vector you passed me is"
                                 f" {len(var_list)} long while cost {var_weight.shape}.")

        if var_ref is not None:
            if isinstance(var_ref, ca.SX) or isinstance(var_ref, ca.MX):
                var_ref_dim = var_ref.shape[0]
            else:
                var_ref_dim = len(var_ref)

            if not len(var_list) == var_ref_dim:
                raise ValueError(f"{type} and reference dimensions must be compatible. The states vector you passed me "
                                 f"is {len(var_list)} long while cost {var_ref_dim}.")

    @staticmethod
    def _create_weight_matrix(arg, name):
        """

        :param arg:
        :param name:
        :return:
        """
        if isinstance(arg, np.ndarray) or isinstance(arg, ca.DM):
            if check_if_square(arg):
                W = arg
            elif arg.ndim == 1:
                W = np.diag(arg)
            else:
                raise TypeError(f"{name} must be a square matrix,a 1-D array or a list of real numbers.")
        elif isinstance(arg, list):
            W = np.diag(arg)
        elif isinstance(arg, float) or isinstance(arg, int):
            W = np.diag([arg])
        else:
            raise TypeError(f"The {name} must be a list of floats, numpy array or casadi DM.")
        return W

    def _check_options(self, trajectory_tracking, path_following, names, references, type):
        """

        :param trajectory_tracking:
        :param path_following:
        :param names:
        :param references:
        :param type:
        :return:
        """
        if trajectory_tracking and path_following:
            raise TypeError("You cannot have both trajectory tracking and path following activated. Please chose one of"
                            " them.")
        elif trajectory_tracking and not path_following:
            self._n_tv_ref += len(names)
            self._has_trajectory_following = True
        elif path_following and not trajectory_tracking:
            if not isinstance(references, ca.SX) and not isinstance(references, ca.MX):
                raise TypeError(f"If path_following=True you need to pass an casadi SX or MX expression to the "
                                f"{type}_ref  parameter.")
            self._has_path_following = True
        elif not path_following and not trajectory_tracking and (
                isinstance(references, ca.SX) or isinstance(references, ca.MX)):
            raise ValueError(f"You passed an {references.type_name()} expression to the {type}_ref. In this case you "
                             f"need to tell me if this is a path following or trajectory tracking problem. \n Please "
                             f"set either 'path_following=True' or trajectory_tracking=True.")
        if check_if_has_duplicates(self.name_varying_trajectories):
            raise TypeError("Two different varying trajectory for the same states are not allowed.")

    def _add_cost_term(self, var, names, W, ref, path_following, trajectory_tracking, ind, type):
        """

        :param var:
        :param names:
        :param W:
        :param ref:
        :param path_following:
        :param trajectory_tracking:
        :param ind:
        :param type:
        :return:
        """
        ref, _ = self._transform_reference(ref)

        self._check_options(trajectory_tracking, path_following, names, ref, type)

        W = self._create_weight_matrix(W, who_am_i())

        self._check_dimensions(names, W, ref, type)

        if ref is None and not trajectory_tracking:
            if type == 'inputs_change':
                if self._use_sx:
                    u_old = ca.SX.sym('_'.join(names) + '_old', var.shape[0])
                else:
                    u_old = ca.MX.sym('_'.join(names) + '_old', var.shape[0])
                self._cost += ca.mtimes((var - u_old).T, ca.mtimes(W, (var - u_old)))
                self.input_change_placeholder = ca.vertcat(self.input_change_placeholder, u_old)
                self._input_change_list.append({'ref': ref, 'names': names, 'placeholder': u_old, 'ind': ind,
                                                'type': type})
            else:
                # Then is a stabilization problem
                self._cost += ca.mtimes(var.T, ca.mtimes(W, var))
        else:
            # Can be either a reference tracking, trajectory tracking or path following problem
            if path_following:
                # If path following, simply substitute the SX/MX expression of the path inside the cost
                p_ref = ref

                # extract the theta variable
                theta = p_ref[0].dep()

                # The paths are saved for plotting
                self._paths_list.append({'ref': ref, 'names': names, 'ind': ind, 'type': type, 'path_fun': None,
                                         'theta': theta, 'ind_theta': None})
            else:
                # For trajectory tracking and reference tracking problems, put a placeholder.
                # This will be scaled and substituted with the reference inside the MPC
                if trajectory_tracking:
                    if self._use_sx:
                        p_ref = ca.SX.sym('_'.join(names) + 'ref', len(names))
                    else:
                        p_ref = ca.MX.sym('_'.join(names) + 'ref', len(names))
                    self.ref_placeholder = ca.vertcat(self.ref_placeholder, p_ref)

                    self._trajectories_list.append({'ref': ref, 'names': names, 'placeholder': p_ref, 'ind': ind,
                                                    'type': type, 'traj_fun': None})
                else:
                    ref = check_and_wrap_to_list(ref)
                    if self._use_sx:
                        p_ref = ca.SX.sym('_'.join(names) + 'ref', len(names))
                    else:
                        p_ref = ca.MX.sym('_'.join(names) + 'ref', len(names))
                    self._references_list.append({'ref': ref, 'names': names, 'placeholder': p_ref, 'ind': ind,
                                                  'type': type})

            self._cost += ca.mtimes((var - p_ref).T, ca.mtimes(W, (var - p_ref)))

        self._is_set = True

    def _setup(self, x_scale=None, u_scale=None, y_scale=None, path_variables=None, time_variable=None):
        """

        :param x_scale:
        :param u_scale:
        :param y_scale:
        :param path_variables:
        :param time_variable:
        :return:
        """
        for i, ref in enumerate(self._references_list):
            # This is for the reference tracking problem. Scale and substitute the real value of the reference.
            ind = ref['ind']
            type = ref['type']
            if type == 'states':
                scale = x_scale
            elif type == 'measurements':
                scale = y_scale
            elif type == 'inputs':
                scale = u_scale
            else:
                raise TypeError(f"Type {type} no available.")

            self._cost = ca.substitute(self._cost, ref['placeholder'], ref['ref'] / ca.DM(scale)[ind])

        for i, traj in enumerate(self._trajectories_list):
            # This is for the trajectory tracking problem. Just scale, the reference will be substituted later.
            ind = traj['ind']
            type = traj['type']

            # Create casadi functions out of the functions used for trajectory-tracking problems. Necessary for plotting
            if isinstance(traj['ref'], ca.SX) or isinstance(traj['ref'], ca.MX):
                traj['traj_fun'] = ca.Function('traj_fun', [time_variable], [traj['ref']])

            if type == 'states':
                scale = x_scale
            elif type == 'measurements':
                scale = y_scale
            elif type == 'inputs':
                scale = u_scale
            else:
                raise TypeError(f"Type {type} no available.")
            self._cost = ca.substitute(self._cost, traj['placeholder'], traj['placeholder'] / ca.DM(scale)[ind])

        # Create casadi functions out of the functions used for path-following problems. Necessary for plotting
        theta_tot = []
        for path_var in path_variables:
            theta_tot.append(path_var['theta'])
        for path in self._paths_list:
            path['path_fun'] = ca.Function('path_fun', [ca.vertcat(*theta_tot)], [path['ref']])

        for i, clist in enumerate(self._input_change_list):
            names = clist['names']
            for name in names:
                indx = self._model.input_names.index(name)
                if indx not in self._input_change_list:
                    self._ind_input_changes.append(indx)

    @staticmethod
    def _transform_reference(ref):
        """

        :param ref:
        :return:
        """
        if isinstance(ref, float) or isinstance(ref, int):
            # Allow the user to input also float or integer as ref. This can be the case if just one variable is passed
            ref = [ref]
        elif check_if_list_of_type(ref, ca.MX) or check_if_list_of_type(ref, ca.SX):
            ref = ca.vertcat(*ref)

        use_sx = True
        if isinstance(ref, ca.MX):
            use_sx = False

        return ref, use_sx

    def add_states(self, names, weights, ref=None, path_following=False, trajectory_tracking=False):
        """

        :param names:
        :param weights:
        :param ref:
        :param path_following:
        :param trajectory_tracking:
        :return:
        """
        names = check_and_wrap_to_list(names)
        ind_x = []
        for i in names:
            try:
                ind_x.append(self._model.dynamical_state_names.index(i))
            except ValueError:
                raise ValueError(f"The state {i} does not exist. The available states are {self._model._x._names}")

        x = self._model._x[ind_x]
        self._add_cost_term(x, names, weights, ref, path_following, trajectory_tracking, ind_x, 'states')

    def add_measurements(self, names, weights, ref=None, path_following=False, trajectory_tracking=False):
        """

        :param names:
        :param weights:
        :param ref:
        :param path_following:
        :param trajectory_tracking:
        :return:
        """
        names = check_and_wrap_to_list(names)
        ind_y = []
        for i in names:
            try:
                ind_y.append(self._model.measurement_names.index(i))
            except ValueError:
                raise ValueError(f"The measurement {i} does not exist. The available measurements are "
                                 f"{self._model._y._names}")

        y = self._model._y[ind_y]
        # Substitute the measurement variables with the measurement equations. These are function of states,inputs or
        # parameters
        m_eq = ca.substitute(y, y, self._model.meas[ind_y])
        self._add_cost_term(m_eq, names, weights, ref, path_following, trajectory_tracking, ind_y, 'measurements')

    def add_inputs(self, names, weights, ref=None, path_following=False, trajectory_tracking=False):
        """

        :param names:
        :param weights:
        :param ref:
        :param path_following:
        :param trajectory_tracking:
        :return:
        """
        names = check_and_wrap_to_list(names)
        ind_u = []
        for i in names:
            try:
                ind_u.append(self._model._u._names.index(i))
            except ValueError:
                raise ValueError(f"The state {i} does not exist. The available states are {self._model._u._names}")

        u = self._model._u[ind_u]
        self._add_cost_term(u, names, weights, ref, path_following, trajectory_tracking, ind_u, 'inputs')

    def add_inputs_change(self, names, weights):
        """

        :param names:
        :param weights:
        :return:
        """
        ref = None
        path_following = False
        trajectory_tracking = False
        names = check_and_wrap_to_list(names)
        ind_u = []
        for i in names:
            try:
                ind_u.append(self._model._u._names.index(i))
            except ValueError:
                raise ValueError(f"The state {i} does not exist. The available states are {self._model._u._names}")

        u = self._model._u[ind_u]
        self._add_cost_term(u, names, weights, ref, path_following, trajectory_tracking, ind_u, 'inputs_change')

    @property
    def n_of_paths(self):
        """
        Number of paths. It's different from zero if path following is active.

        :return:
        """
        return len(self._paths_list)

    @property
    def n_of_trajectories(self):
        """
        Number of paths. It's different from zero if path following is active.

        :return:
        """
        return len(self._trajectories_list)

    @property
    def name_varying_trajectories(self):
        """

        :return:
        """
        for l in self._trajectories_list:
            self._varying_trajectories_names.extend(l['names'])
        return self._varying_trajectories_names

    @property
    def name_open_varying_trajectories(self):
        """

        :return:
        """
        for l in self._trajectories_list:
            if l['ref'] is None:
                self._open_varying_trajectories_names.extend(l['names'])
        return self._open_varying_trajectories_names

    @property
    def Q(self):
        """
        :return:
        """
        if self._Q is None:
            self._Q = ca.DM(ca.jacobian(ca.jacobian(self.cost, self._model.x), self._model.x))/2
        return self._Q

    @Q.setter
    def Q(self, arg):
        self._Q = arg

    @property
    def R(self):
        """

        :return:
        """
        if self._R is None:
            self._R = ca.DM(ca.jacobian(ca.jacobian(self.cost, self._model.u), self._model.u))/2
        return self._R

    @R.setter
    def R(self, arg):
        self._R = arg

    @property
    def P(self):
        """

        :return:
        """
        return self._P

    @P.setter
    def P(self, arg):
        if self._P is None:
            self._P = ca.DM(ca.jacobian(ca.jacobian(self.cost, self._model.x), self._model.x))/2
        self._P = arg


class MHEQuadraticCost(GenericCost):
    """"""

    def __init__(self, model, use_sx=True):
        """Constructor method"""
        super().__init__(model, use_sx)

        # List of references that have a fixed values for all MHE iterations
        self._fixed_ref_list = []

        # List of references that change at every time step of the MHE horizon
        self._tv_ref_list = []

        # List of references that change every at every iteration, but not withing the MHE horizon:
        self._iter_ref_list = []

        # Placeholders for the references, needed for trajectory-tracking problems
        self.ref_placeholder = []
        self._has_state_noise = False
        self._w = []
        self._x_guess = None
        self._p_guess = None
        self._n_tv_refs = 0
        self._n_iter_var_refs = 0
        self._name_long_to_short = {'states': 'x', 'inputs': 'u', 'measurements': 'y', 'parameters': 'p',
                                    'state_noise': 'w'}

    @staticmethod
    def _check_dimensions(var_list, var_weight, var_ref, type):
        """
        Run some sanity checks on the user supplied values

        :return:
        """
        if var_list is not None and var_weight is None:
            raise ValueError(
                f"You passed the following {type}: {var_list} to the cost function, "
                f"but I do not have any weights for it/them. "
                f"Please pass me the weights."
            )

        if var_list is None and var_weight is not None:
            raise ValueError(f"You passed the {type} weights {var_weight} but without state names. "
                             f"Please pass me the state names with quad_stage_cost.x.")

        if not check_if_list_of_none([var_list, var_weight]):
            if not len(var_list) == var_weight.shape[0]:
                raise ValueError(f"{type} and weights dimensions must be compatible."
                                 f"The {type} vector you passed me is {len(var_list)} long "
                                 f"while cost {var_weight.shape}.")

        if var_ref is not None:
            if isinstance(var_ref, ca.SX):
                var_ref_dim = var_ref.shape[0]
            else:
                var_ref_dim = len(var_ref)

            if not len(var_list) == var_ref_dim:
                raise ValueError(f"{type} and reference dimensions must be compatible."
                                 f"The states vector you passed me is {len(var_list)} long while cost {var_ref_dim}.")

    @staticmethod
    def _create_weight_matrix(arg, name):
        """

        :param arg:
        :param name:
        :return:
        """
        if isinstance(arg, np.ndarray) or isinstance(arg, ca.DM):
            if check_if_square(arg):
                W = arg
            elif arg.ndim == 1:
                W = np.diag(arg)
            else:
                raise TypeError(f"{name} must be a square matrix, a 1-D array or a list of real numbers.")
        elif isinstance(arg, list):
            W = np.diag(arg)
        elif isinstance(arg, float) or isinstance(arg, int):
            W = np.diag([arg])
        else:
            raise TypeError(f"The {name} must be a list of floats, numpy array or CasADi DM.")
        return W

    def _add_cost_term(self, var, names, W, ref, ind, type, ref_type=None):
        """

        :param var:
        :param names:
        :param W:
        :param ref:
        :param ind:
        :param type:
        :param ref_type:
        :return:
        """
        name = self._name_long_to_short[type]
        if isinstance(ref, float) or isinstance(ref, int):
            # Allow the user to input also float or integer as ref. This can be the case if just one variable is passed
            ref = [ref]

        W = self._create_weight_matrix(W, who_am_i())

        self._check_dimensions(names, W, ref, type)

        if ref is None:
            # Then is a stabilization problem
            self._cost += ca.mtimes(var.T, ca.mtimes(W, var))
        else:
            if ref_type == 'tv_ref':
                # Time varying references. These reference change within the prediction horizon, i.e. they are function
                # of the states and inputs
                p_ref = ca.SX.sym(f'{name}_ref', ref.shape[0])
                self._n_tv_refs += ref.shape[0]
                self._tv_ref_list.append({'ref': ref, 'names': names, 'placeholder': p_ref, 'ind': ind, 'type': type})
            elif ref_type == 'iter_var_ref':
                # Iteration varying references. These reference change only once for every MHE iteration
                ref = check_and_wrap_to_list(ref)
                p_ref = ca.SX.sym(f'{name}_ref', len(ref))
                self._n_iter_var_refs += len(ref)
                self._iter_ref_list.append({'ref': ref, 'names': names, 'placeholder': p_ref, 'ind': ind, 'type': type})
            elif ref_type == 'fixed_ref':
                # Fixed references. The values of these reference do not change within the prediction horizon
                ref = check_and_wrap_to_list(ref)
                p_ref = ca.SX.sym(f'{name}_ref', len(ref))
                self._fixed_ref_list.append(
                    {'ref': ref, 'names': names, 'placeholder': p_ref, 'ind': ind, 'type': type}
                )
            else:
                raise ValueError(f"{ref_type} is not known.")

            self._cost += ca.mtimes((var - p_ref).T, ca.mtimes(W, (var - p_ref)))

        self._is_set = True

    def _setup(self, x_scale=None, w_scale=None, p_scale=None, u_scale=None):
        """
        This scales the references of path following and trajectory tracking.

        :param x_scale:
        :param w_scale:
        :param p_scale:
        :param u_scale:
        :return:
        """
        for fixed_ref in self._fixed_ref_list:
            self._cost = ca.substitute(self._cost, fixed_ref['placeholder'], fixed_ref['ref'])
        self._cost = ca.substitute(self._cost, self._model.x, self._model.x * ca.DM(x_scale))
        if self._model.n_p > 0:
            self._cost = ca.substitute(self._cost, self._model.p, self._model.p * ca.DM(p_scale))
        if self._has_state_noise:
            self._cost = ca.substitute(self._cost, self._w, self._w * ca.DM(w_scale))

    def add_measurements(self, weights, names=None):
        """

        :param weights:
        :param names:
        :return:
        """
        if names is None:
            names = self._model.measurement_names
            ind_y = list(range(self._model.n_y))
        else:
            names = check_and_wrap_to_list(names)
            ind_y = []
            for i in names:
                try:
                    ind_y.append(self._model._y._names.index(i))
                except ValueError:
                    raise ValueError(
                        f"The measurement {i} does not exist. The available states are {self._model._y._names}"
                    )

        y = self._model.y[ind_y]

        # Substitute the measurement variables with the measurement equations.
        # These are function of states,inputs or parameters
        m_eq = ca.substitute(y, y, self._model.meas[ind_y])
        self._add_cost_term(m_eq, names, weights, y, ind_y, 'measurements', 'tv_ref')

    def add_inputs(self, names, weights):
        """

        :param names:
        :param weights:
        :return:
        """
        names = check_and_wrap_to_list(names)
        ind_u = []
        for i in names:
            try:
                ind_u.append(self._model._u._names.index(i))
            except ValueError:
                raise ValueError(
                    f"The state {i} does not exist. The available states are {self._model._u._names}"
                )

        u = self._model.u[ind_u]
        ref = ca.SX.sym('u_meas', len(ind_u))
        self._add_cost_term(u, names, weights, ref, ind_u, 'inputs', 'tv_ref')

    def add_state_noise(self, weights):
        """

        :param weights:
        :return:
        """
        self._w = ca.SX.sym('w', self._model.n_x)
        names = [f'w_{i}' for i in self._model.dynamical_state_names]
        ind_w = list(range(self._model.n_x))
        self._has_state_noise = True
        self._add_cost_term(self._w, names, weights, None, ind_w, 'state_noise')

    def add_states(self, weights, guess):
        """

        :param weights:
        :param guess:
        :return:
        """
        names = self._model.dynamical_state_names
        guess = check_and_wrap_to_list(guess)
        if len(guess) != self._model.n_x:
            raise ValueError(f"The guess must have the same dimension of the model states."
                             f"There are {self._model.n_x} states but guess has {len(guess)} values.")
        ind_x = list(range(self._model.n_x))
        self.x_guess = guess
        self._add_cost_term(self._model.x, names, weights, guess, ind_x, 'states', 'iter_var_ref')

    def add_parameters(self, weights, guess):
        """

        :param weights:
        :param guess:
        :return:
        """
        names = self._model.parameter_names
        ind_p = list(range(self._model.n_p))
        guess = check_and_wrap_to_list(guess)
        if len(guess) != self._model.n_p:
            raise ValueError(f"The guess must have the same dimension of the model parameters."
                             f"There are {self._model.n_p} states but guess has {len(guess)} values.")
        self.p_guess = guess
        self._add_cost_term(self._model.p, names, weights, guess, ind_p, 'parameters', 'iter_var_ref')

    @property
    def w(self):
        """

        :return:
        """
        return self._w

    @property
    def x_guess(self):
        """

        :return:
        """
        return self._x_guess

    @x_guess.setter
    def x_guess(self, arg):
        self._x_guess = arg

    @property
    def p_guess(self):
        """

        :return:
        """
        return self._p_guess

    @p_guess.setter
    def p_guess(self, arg):
        self._p_guess = arg

    @property
    def n_iter_var_refs(self):
        """

        :return:
        """
        return self._n_iter_var_refs


class GenericConstraint:
    """Class for generic constraints"""

    def __init__(self, model, name='constraint'):
        """Constructor method"""
        self._is_soft = False
        self._ub = None
        self._lb = None
        self._function = None
        self._name = name
        self._formatted_name = name.replace(' ', '_')
        self._model = model
        self._is_set = False
        self._size = 0
        self._cost = None
        self._weight = None
        self.e_soft_value = 0
        self._max_violation = ca.inf

    def _check_and_setup(self, x_scale=None, u_scale=None, y_scale=None, p_scale=None):
        """

        :param x_scale:
        :param u_scale:
        :param y_scale:
        :param p_scale:
        :return:
        """
        if self.constraint is not None:
            for i in range(self._model.n_x):
                self.constraint = ca.substitute(self.constraint, self._model.x[i], self._model.x[i] * x_scale[i])

            for i in range(self._model.n_u):
                self.constraint = ca.substitute(self.constraint, self._model.u[i], self._model.u[i] * u_scale[i])

            for i in range(self._model.n_y):
                self.constraint = ca.substitute(self.constraint, self._model.y[i], self._model.y[i] * y_scale[i])

        if self.is_set:
            if self.size != len(self._ub):
                raise ValueError("The dimensions of the terminal constraint function and its upper bound are not "
                                 "compatible. The terminal constraint must have dimension (n x 1) where n is the number"
                                 " of upper bounds.")
            if self.size != len(self._lb):
                raise ValueError("The dimensions of the terminal constraint function and its lower bound are not "
                                 "compatible. The terminal constraint must have dimension (n x 1) where n is the lower "
                                 "of upper bounds.")
            if self.is_soft:
                e = ca.SX.sym('e', self.size)
                self._cost = ca.Function(f'cost_{self._formatted_name}', [e],
                                         [ca.mtimes(ca.mtimes(e.T, self._weight), e)])
        else:
            if self._function is not None:
                if self.is_soft:
                    if self._weight is None:
                        self._weight = np.diag(np.ones(self.size) * 10000)
                    e = ca.SX.sym('e', self.size)
                    self._cost = ca.Function(f'cost_{self._formatted_name}', [e],
                                             [ca.mtimes(ca.mtimes(e.T, self._weight), e)])

                if self.lb is None:
                    self.lb = -np.inf * np.ones(self._size)

                if self.ub is None:
                    self.ub = np.inf * np.ones(self._size)

                self._is_set = True

    @property
    def cost(self):
        """

        :return:
        """
        if self.is_set:
            return self._cost

    @property
    def constraint(self):
        """

        :return:
        """
        return self._function

    @constraint.setter
    def constraint(self, arg):
        if arg is not None:
            if not isinstance(arg, ca.SX) and not isinstance(arg, ca.MX):
                raise TypeError(f"The {self._name} must be of type casadi SX/MX or None.")
        self._function = arg

    @property
    def ub(self):
        """

        :return:
        """
        return self._ub

    @ub.setter
    def ub(self, arg):
        if arg is not None:
            arg = check_and_wrap_to_list(arg)
        self._ub = arg

    @property
    def lb(self):
        """

        :return:
        """
        return self._lb

    @lb.setter
    def lb(self, arg):
        if arg is not None:
            arg = check_and_wrap_to_list(arg)
        self._lb = arg

    @property
    def is_soft(self):
        """

        :return:
        """
        return self._is_soft

    @is_soft.setter
    def is_soft(self, arg):
        if not isinstance(arg, bool):
            raise TypeError("is_soft must be of type bool")
        self._is_soft = arg

    @property
    def max_violation(self):
        """

        :return:
        """
        return self._max_violation

    @max_violation.setter
    def max_violation(self, arg):
        if arg is not None:
            arg = check_and_wrap_to_list(arg)
        self._max_violation = arg

    @property
    def is_set(self):
        """

        :return:
        """
        if self.is_soft:
            if all(v is not None for v in [self._weight, self._function, self._ub, self._lb]):
                return True
            else:
                return False
        else:
            if all(v is not None for v in [self._function, self._ub, self._lb]):
                return True
            else:
                return False

    @property
    def size(self):
        """

        :return:
        """
        if self.constraint is not None:
            self._size = self.constraint.size1()
        return self._size

    @property
    def weight(self):
        """

        :return:
        """
        return self._weight

    @weight.setter
    def weight(self, arg):
        self._weight = arg


EXPLICIT_METHODS = {
    'forward_euler': {
        'A': np.array([[0.]]),
        'b': np.array([1.]),
        'c': np.array([0.])
    },
    'midpoint': {
        'A': np.array([[0., 0.],
                       [.5, 0.]]),
        'b': np.array([0., 1.]),
        'c': np.array([0., .5])
    },
    'heun': {
        '2': {
            'A': np.array([[0., 0.],
                           [1., 0.]]),
            'b': np.array([.5, .5]),
            'c': np.array([0., 1.])
        },
        '3': {
            'A': np.array([[0., 0., 0.],
                           [1 / 3, 0., 0.],
                           [0., 2 / 3, 0.]]),
            'b': np.array([.25, 0., .75]),
            'c': np.array([0., 1 / 3, 2 / 3])
        }
    },
    'ralston': {
        '2': {
            'A': np.array([[0., 0.],
                           [2 / 3, 0.]]),
            'b': np.array([.25, .75]),
            'c': np.array([0., 2 / 3])
        },
        '3': {
            'A': np.array([[0., 0., 0.],
                           [.5, 0., 0.],
                           [0., .75, 0.]]),
            'b': np.array([2 / 9, 1 / 3, 4 / 9]),
            'c': np.array([0., .5, .75])
        },
        '4': {}
    },
    'generic': {
        '2': {},
        '3': {}
    },
    'kutta': {
        'A': np.array([[0., 0., 0.],
                       [.5, 0., 0.],
                       [-1., 2., 0.]]),
        'b': np.array([1 / 6, 2 / 3, 1 / 6]),
        'c': np.array([0., .5, 1.])
    },
    'ssprk3': {
        'A': np.array([[0., 0., 0.],
                       [1., 0., 0.],
                       [.25, .25, 0.]]),
        'b': np.array([1 / 6, 1 / 6, 2 / 3]),
        'c': np.array([0., 1., .5])
    },
    'classic': {
        'A': np.array([[0., 0., 0., 0.],
                       [.5, 0., 0., 0.],
                       [0., .5, 0., 0.],
                       [0., 0., 1., 0.]]),
        'b': np.array([1 / 6, 1 / 3, 1 / 3, 1 / 6]),
        'c': np.array([0., .5, .5, 1.])
    },
    '3_8_rule': {
        'A': np.array([[0., 0., 0., 0.],
                       [1 / 3, 0., 0., 0.],
                       [-1 / 3, 0., 0., 0.],
                       [1., -1., 1., 0.]]),
        'b': np.array([1 / 8, 3 / 8, 3 / 8, 1 / 8]),
        'c': np.array([0., 1 / 3, 2 / 3, 1.])
    }
}


class RungeKutta:
    """"""

    @staticmethod
    def _construct_polynomial_basis(
            degree: int,
            method: str,
            h: Symbolic
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """

        :param degree:
        :param method:
        :return:
        """
        B = np.zeros(degree + 1)
        C = np.zeros((degree + 1, degree + 1))
        D = np.zeros(degree + 1)
        # T = np.zeros(degree + 1)
        T = h.zeros(degree + 1)
        tau = [0.] + ca.collocation_points(degree, method)

        for i in range(degree + 1):
            L = np.poly1d([1.])
            for j in range(degree + 1):
                if j != i:
                    L *= np.poly1d([1., -tau[j]]) / (tau[i] - tau[j])

            D[i] = L(1.)

            Ldot = np.polyder(L)
            for j in range(degree + 1):
                C[i, j] = Ldot(tau[j])

            Lint = np.polyint(L)
            B[i] = Lint(1.)

            T[i] = h * tau[i]

        return B, C, D, T

    @classmethod
    def _collocation(cls, problem, **opts):
        """

        :param problem:
        :param opts:
        :return:
        """
        # TODO: Add support for algebraic variables (DAE)
        ode = problem.get('ode')
        alg = problem.get('alg')
        quad = problem.get('quad')
        meas = problem.get('meas')
        t = problem.get('t')
        x = problem.get('x')
        z = problem.get('z')
        u = problem.get('u')
        p = problem.get('p')
        dt = problem.get('dt')

        if meas is not None:
            if not isinstance(meas, (ca.SX, ca.MX)):
                # TODO: What is this supposed to do?
                if not meas.is_empty():
                    meas = ca.Function('meas', [x, z, u, p], [meas])
                else:
                    meas = ca.SX()

        degree = opts.get('degree', None)
        collocation_points = opts.get('collocation_points', None)
        if degree is None:
            degree = 2
        if collocation_points is None:
            collocation_points = 'radau'
        n_x = x.size1()
        n_z = z.size1()
        function = ca.Function('function', [t, x, z, u, p], [ode, alg, quad], {'allow_free': True})

        B, C, D, T = cls._construct_polynomial_basis(degree, collocation_points, dt)  # h instead of dt

        J = 0.
        ce = []

        xc = []
        zc = []  # not sure about this
        for k in range(degree):
            xk = x.sym('x_' + str(k), n_x)
            xc.append(xk)
            zk = z.sym('z_' + str(k), n_z)  # not sure about this
            zc.append(zk)  # not sure about this

        xf = D[0] * x
        zf = 0.  # not sure about this
        for i in range(1, degree + 1):
            xp = C[0, i] * x
            for j in range(degree):
                xp += C[j + 1, i] * xc[j]

            fi, gi, qi = function(t + T[i - 1], xc[i - 1], zc[i - 1], u, p)  # not sure about this
            # (zc[i - 1] instead of z)
            ce.append(dt * fi - xp)  # h instead of dt
            ce.append(gi)  # not sure about this

            xf += D[i] * xc[i - 1]
            zf += D[i] * zc[i - 1]  # not sure about this

            J += B[i] * qi * dt  # h instead of dt

        # xc = ca.vertcat(*xc)
        # zc = ca.vertcat(*zc)
        ce = ca.vertcat(*ce)

        if meas is not None:
            if isinstance(meas, ca.Function):
                meas = meas(xf, zf, u, p)  # not sure about this (zf instead of z)

        problem['ode'] = xf
        problem['alg'] = zf  # not sure about this (zf instead of alg)
        problem['quad'] = J
        problem['meas'] = meas
        problem['collocation_points_ode'] = xc
        problem['collocation_points_alg'] = zc  # not sure about this
        problem['collocation_equations'] = ce

    @classmethod
    def _explicit(cls, problem, **opts):
        """

        :param problem:
        :param opts:
        :return:
        """
        # TODO: Integrate algebraic variables
        # TODO: Deal with None values
        ode = problem.get('ode')
        alg = problem.get('alg')
        quad = problem.get('quad')
        h = problem.get('dt')
        t = problem.get('t')
        x = problem.get('x')
        z = problem.get('z')
        u = problem.get('u')
        p = problem.get('p')

        order = opts.get('order')
        if order is None:
            order = 1
        dyn = ca.Function('dyn', [t, x, z, u, p], [ode, quad], {'allow_free': True})
        alg = ca.Function('alg', [t, x, z, u, p], [alg], {'allow_free': True})
        butcher_tableau = opts.get('butcher_tableau', None)
        if butcher_tableau is None:
            if order == 1:
                butcher_tableau = 'forward_euler'
            elif order == 2:
                butcher_tableau = 'midpoint'
            elif order == 3:
                butcher_tableau = 'kutta'
            elif order == 4:
                butcher_tableau = 'classic'
            else:
                raise NotImplementedError(f"Explicit Runge-Kutta discretization for order {order} is not yet "
                                          f"implemented.")
        if butcher_tableau in ['heun', 'ralston', 'generic']:
            butcher_tableau = EXPLICIT_METHODS[butcher_tableau][str(order)]
        else:
            butcher_tableau = EXPLICIT_METHODS[butcher_tableau]
        A = butcher_tableau['A']
        b = butcher_tableau['b']
        c = butcher_tableau['c']
        k = []
        ka = []
        kq = []
        Z = z.sym('Z', (z.size1(), order))
        for i in range(order):
            ki = 0
            for j in range(i):
                ki += A[i, j] * k[j]
            dk, dkq = dyn(t + h * c[i], x + h * ki, Z[:, i], u, p)
            k.append(dk)
            ka.append(alg(t + h * c[i], dk, Z[:, i], u, p))
            kq.append(dkq)

        ode = x
        quad = 0
        for i in range(order):
            ode += h * b[i] * k[i]
            quad += h * b[i] * kq[i]
        alg = ca.vertcat(*ka)

        problem['discretization_points'] = Z[:]
        problem['ode'] = ode
        problem['alg'] = alg
        problem['quad'] = quad

    @classmethod
    def discretize(cls, problem, **opts):
        """

        :param problem:
        :param opts:
        :return:
        """
        class_ = opts.pop('class', 'explicit')
        method = opts.pop('method', 'rk')

        if class_ == 'explicit':
            if method == 'rk':
                cls._explicit(problem, **opts)
        elif class_ == 'implicit':
            if method == 'collocation':
                cls._collocation(problem, **opts)
        else:
            raise ValueError(f"Runge-Kutta class {class_} not recognized")


def continuous2discrete(problem, category='runge-kutta', **opts):
    """

    :param problem:
    :param category:
    :param opts:
    :return:
    """
    # TODO: Maybe add inplace feature? (If inplace=False, then a new dictionary will be returned)
    if category == 'runge-kutta':
        RungeKutta.discretize(problem, **opts)






#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from abc import ABCMeta, abstractmethod
from typing import Optional, Union
import warnings

import casadi as ca
import numpy as np

from .base import _Estimator
from ..dynamic_model.dynamic_model import Model


class _KalmanFilter(_Estimator, metaclass=ABCMeta):
    """
    Kalman filter base class

    :param model:
    :param id: The identifier of the KF object. If no identifier is given, a random one will be generated.
    :param name: The name of the KF object. By default the KF object has no name.
    :param plot_backend: Plotting library that is used to visualize estimated data. At the moment only
        `Matplotlib <https://matplotlib.org/>`_ and `Bokeh <https://bokeh.org/>`_ are supported. By default, no plotting
        library is selected, i.e. no plots can be generated.
    :param square_root_form: Not used at the moment (will be implemented in the future)
    """
    def __init__(
            self,
            model: Model,
            id: Optional[str] = None,
            name: Optional[str] = None,
            plot_backend: Optional[str] = None,
            square_root_form: bool = True
    ) -> None:
        """Constructor method"""
        super().__init__(model, id=id, name=name, plot_backend=plot_backend)

        self._square_root_form = square_root_form
        self._predict_function = None
        self._update_function = None
        self._is_linearized = False

    @abstractmethod
    def _setup_parameters(self) -> None:
        """

        :return:
        """
        pass

    def _setup_predict(self, *args: Optional[ca.Function]) -> None:
        """

        :param args:
        :return:
        """
        state_matrix = args[0]

        n_x = self._model.n_x
        n_u = self._model.n_u
        n_p = self._model.n_p

        x = ca.MX.sym('x', n_x)
        u = ca.MX.sym('u', n_u)
        p = ca.MX.sym('p', n_p)

        P = ca.SX.sym('P', (n_x, n_x))
        Q = ca.SX.sym('Q', (n_x, n_x))
        if self._is_linearized:
            # NOTE: For now we ignore time-variant systems (i.e., [] instead of t)
            F = state_matrix(self._model.x, [], self._model.u, self._model.p, self._solution.dt, [])
        else:
            # NOTE: For now we ignore time-variant systems (i.e., [] instead of t)
            F = state_matrix(self._model.p, self._solution.dt, [])
        if self._model.discrete:
            Pdot = F @ P @ F.T + Q
        else:
            Pdot = F @ P + P @ F.T + Q

        model = self._model.copy(setup=False)

        model.add_dynamical_states(P[:])
        model.add_dynamical_equations(Pdot[:])
        model.add_parameters(Q[:])
        if self._solution.dt is not None:
            model.setup(dt=self._solution.dt)
        elif self._solution.grid is not None:
            model.setup(grid=self._solution.grid)
        else:
            model.setup()

        self._solution.setup('P', P={
            'values_or_names': [k.name() for k in P.elements()],
            'description': P.numel() * [''],
            'labels': P.numel() * [''],
            'units': P.numel() * [''],
            'shape': (P.numel(), 0),
            'data_format': ca.DM
        })

        P = ca.MX.sym('P', P.shape)
        Q = ca.MX.sym('Q', Q.shape)

        sol = model(x0=ca.vertcat(x, P[:]), p=ca.vertcat(u, p, Q[:]))
        result = sol['xf']
        x_pred = result[:n_x]
        P_pred = ca.reshape(result[n_x:], n_x, n_x)

        self._predict_function = ca.Function('prediction_step',
                                             [ca.horzcat(x, P), ca.vertcat(u, p), Q],
                                             [ca.horzcat(x_pred, P_pred)],
                                             ['x0', 'p', 'Q'],
                                             ['x'])

    def _setup_update(self, *args: Optional[ca.Function]) -> None:
        """

        :param args:
        :return:
        """
        output_matrix = args[0]

        n_x = self._model.n_x
        n_y = self._model.n_y
        n_u = self._model.n_u
        n_p = self._model.n_p

        x = ca.MX.sym('x', n_x)
        u = ca.MX.sym('u', n_u)
        p = ca.MX.sym('p', n_p)
        up = ca.vertcat(u, p)
        P = ca.MX.sym('P', (n_x, n_x))

        if n_y == 0:
            y_pred = x
            n_y = n_x
        else:
            sol = self._model(x0=x, p=up, which='meas_function')
            y_pred = sol['yf']
        y = ca.MX.sym('y', n_y)

        if self._is_linearized:
            # NOTE: For now we ignore time-variant systems (i.e., [] instead of t)
            H = output_matrix(x, [], u, p, self._solution.dt, [])
        else:
            # NOTE: For now we ignore time-variant systems (i.e., [] instead of t)
            H = output_matrix(p, self._solution.dt, [])
        R = ca.MX.sym('R', (n_y, n_y))
        P_xy = P @ H.T
        P_yy = H @ P @ H.T + R
        # NOTE: Since matrices cannot really be divided (the "/"-operator divides matrices element-wise), a linear
        #  system of equations needs to be solved (solve() in CasADi, which corresponds to the "\"-operator in MATLAB)
        # K = P_xy*P_yy^-1
        # K*P_yy = P_xy
        # P_yy^T*K^T = P_xy^T
        # K^T = P_yy^T\P_xy^T
        K = ca.solve(P_yy.T, P_xy.T).T

        x_up = x + K @ (y - y_pred)
        P_up = P - K @ P_yy @ K.T

        self._update_function = ca.Function('update_step',
                                            [ca.horzcat(x, P), y, up, R],
                                            [ca.horzcat(x_up, P_up), y_pred],
                                            ['x0', 'y', 'p', 'R'],
                                            ['x', 'y_pred'])

    def check_consistency(self) -> None:
        """

        :return:
        """
        super().check_consistency()

        if self._predict_function is not None:
            if self._predict_function.has_free():
                free_vars = ", ".join(self._predict_function.get_free())
                msg = f"Instance {self.__class__.__name__} has the following free variables/parameters: {free_vars}"
                raise RuntimeError(msg)

        if self._update_function is not None:
            if self._update_function.has_free():
                free_vars = ", ".join(self._update_function.get_free())
                msg = f"Instance {self.__class__.__name__} has the following free variables/parameters: {free_vars}"
                raise RuntimeError(msg)

    def setup(self, **kwargs) -> None:
        """

        :param kwargs:
        :return:
        """
        self._solution.setup(self._model.solution)

        if not self._model.is_linear() and self.type == 'extended Kalman filter':
            model = self._model.linearize()
            self._is_linearized = True
        else:
            model = self._model.copy(setup=False)
            self._is_linearized = False

        n_y = model.n_y
        if n_y == 0:
            warnings.warn(f"The model has no measurement equations, I am assuming measurements of all states "
                          f"{self._model.dynamical_state_names} are available.")
            model.set_measurement_equations(model.dynamical_states)

        if self.type != 'unscented Kalman filter':
            arg_in = [model.p, model.dt, model.t]
            if self._is_linearized:
                arg_in = [model.x_eq, model.z_eq, model.u_eq] + arg_in
            state_matrix = ca.Function('state_matrix', arg_in, [model.A])
            output_matrix = ca.Function('output_matrix', arg_in, [model.C])
            predict_args = (state_matrix,)
            update_args = (output_matrix,)
        else:
            predict_args = tuple()
            update_args = tuple()

        n_x = self._model.n_x
        n_y = self._model.n_y
        n_u = self._model.n_u
        n_p = self._model.n_p

        x = ca.MX.sym('x', n_x)
        y = ca.MX.sym('y', n_y)
        u = ca.MX.sym('u', n_u)
        p = ca.MX.sym('p', n_p)
        up = ca.vertcat(u, p)
        P = ca.MX.sym('P', (n_x, n_x))
        Q = ca.MX.sym('Q', (n_x, n_x))
        R = ca.MX.sym('R', (n_y, n_y))

        self._setup_parameters()
        self._setup_update(*update_args)
        self._setup_predict(*predict_args)

        prediction = self._predict_function(ca.horzcat(x, P), up, Q)
        update, y_pred = self._update_function(prediction, y, up, R)

        self._function = ca.Function('function',
                                     [ca.horzcat(x, P), y, up, Q, R],
                                     [update, y_pred],
                                     ['x0', 'y', 'p', 'Q', 'R'],
                                     ['x', 'y_pred'])

        self.check_consistency()

        self._n_x = n_x
        self._n_y = n_y
        # self._n_z = n_z
        self._n_u = n_u
        self._n_p = n_p
        # self._n_p_est = n_p_est

        self._process_noise_covariance = ca.DM.zeros(Q.shape)
        self._measurement_noise_covariance = ca.DM.zeros(R.shape)

    def estimate(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        # TODO: Adjust according to Model.simulate
        self._check_setup()

        args = self._process_inputs(**kwargs)
        tf = args.pop('t0')
        steps = args.pop('steps')
        # TODO: Maybe create a TimeSeries specifically tailored to estimation, where get_function_args(.) returns last
        #  error covariance as well
        args['x0'] = ca.horzcat(args['x0'], ca.reshape(self._solution.get_by_id('P:f'), self._n_x, self._n_x))

        if steps > 1:
            Q = ca.repmat(self._process_noise_covariance, 1, steps)
            R = ca.repmat(self._measurement_noise_covariance, 1, steps)
            args['Q'] = Q
            args['R'] = R
            function = self._function.mapaccum(steps)
            result = function(**args)
        else:
            args['Q'] = self._process_noise_covariance
            args['R'] = self._measurement_noise_covariance
            result = self._function(**args)
        self._solution.update(t=tf, x=result['x'][:, 0], P=result['x'][:, 1:][:], y=result['y_pred'])

    def predict(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        return self._predict_function(*args, **kwargs)

    def update(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        return self._update_function(*args, **kwargs)


class KalmanFilter(_KalmanFilter):
    """
    Kalman filter (KF) class for state estimation (parameter estimation will follow soon)

    :param model:
    :param id: The identifier of the KF object. If no identifier is given, a random one will be generated.
    :param name: The name of the KF object. By default the KF object has no name.
    :param plot_backend: Plotting library that is used to visualize estimated data. At the moment only
        `Matplotlib <https://matplotlib.org/>`_ and `Bokeh <https://bokeh.org/>`_ are supported. By default no plotting
        library is selected, i.e. no plots can be generated.
    :param square_root_form: Not used at the moment (will be implemented in the future)
    """
    def __init__(
            self,
            model: Model,
            id: Optional[str] = None,
            name: Optional[str] = None,
            plot_backend: Optional[str] = None,
            square_root_form: bool = True
    ) -> None:
        """Constructor method"""
        if not model.is_linear():
            raise ValueError("The supplied model is nonlinear. Please use an estimator targeted at the estimation of "
                             "nonlinear systems.")

        super().__init__(model, id=id, name=name, plot_backend=plot_backend, square_root_form=square_root_form)

    def _update_type(self) -> None:
        """

        :return:
        """
        self._type = 'Kalman filter'

    def _setup_parameters(self) -> None:
        """

        :return:
        """
        pass


class ExtendedKalmanFilter(_KalmanFilter):
    """
    Extended Kalman filter (EKF) class for state estimation (parameter estimation will follow soon)

    :param model:
    :param id: The identifier of the EKF object. If no identifier is given, a random one will be generated.
    :param name: The name of the EKF object. By default the EKF object has no name.
    :param plot_backend: Plotting library that is used to visualize estimated data. At the moment only
        `Matplotlib <https://matplotlib.org/>`_ and `Bokeh <https://bokeh.org/>`_ are supported. By default no plotting
        library is selected, i.e. no plots can be generated.
    :param square_root_form: Not used at the moment (will be implemented in the future)
    :note: The same methods and properties as for the :py:class:`Kalman filter <.KalmanFilter>` apply
    """
    def __init__(
            self,
            model: Model,
            id: Optional[str] = None,
            name: Optional[str] = None,
            plot_backend: Optional[str] = None,
            square_root_form: bool = True
    ) -> None:
        """Constructor method"""
        if model.is_linear():
            warnings.warn("The supplied model is linear. For better efficiency use an observer targeted at the "
                          "estimation of linear systems.")

        super().__init__(model, id=id, name=name, plot_backend=plot_backend, square_root_form=square_root_form)

    def _update_type(self) -> None:
        """

        :return:
        """
        self._type = "extended Kalman filter"

    def _setup_parameters(self) -> None:
        """

        :return:
        """
        pass


class UnscentedKalmanFilter(_KalmanFilter):
    """
    Unscented Kalman filter (UKF) class for state estimation (parameter estimation will follow soon)

    :param model:
    :param id: The identifier of the UKF object. If no identifier is given, a random one will be generated.
    :param name: The name of the UKF object. By default the UKF object has no name.
    :param alpha:
    :param beta:
    :param kappa:
    :param plot_backend: Plotting library that is used to visualize estimated data. At the moment only
        `Matplotlib <https://matplotlib.org/>`_ and `Bokeh <https://bokeh.org/>`_ are supported. By default no plotting
        library is selected, i.e. no plots can be generated.
    :param square_root_form: Not used at the moment (will be implemented in the future)
    """
    def __init__(
            self,
            model: Model,
            id: Optional[str] = None,
            name: Optional[str] = None,
            alpha: Union[int, float] = None,
            beta: Union[int, float] = None,
            kappa: Union[int, float] = None,
            plot_backend: Optional[str] = None,
            square_root_form: bool = True
    ) -> None:
        """Constructor method"""
        if model.is_linear():
            warnings.warn("The supplied model is linear. For better efficiency use an observer targeted at the "
                          "estimation of linear systems.")

        super().__init__(model, id=id, name=name, plot_backend=plot_backend, square_root_form=square_root_form)

        if alpha is None:
            alpha = .001
        self._check_parameter_bounds('alpha', alpha)
        if beta is None:
            beta = 2.
        self._check_parameter_bounds('beta', beta)
        if kappa is None:
            kappa = 0.
        self._check_parameter_bounds('kappa', kappa)
        self._lambda = None
        self._gamma = None

        self._weights = None
        self._sqrt = None

    def _update_type(self) -> None:
        """

        :return:
        """
        self._type = "unscented Kalman filter"

    def _check_parameter_bounds(self, param: str, value: Union[int, float]) -> None:
        """

        :param param:
        :param value:
        :return:
        """
        if param == 'alpha':
            if value <= 0. or value > 1.:
                raise ValueError(f"The parameter alpha needs to lie in the interval (0, 1]. Supplied alpha is {value}.")
            self._alpha = value
        elif param == 'beta':
            self._beta = value
        elif param == 'kappa':
            if value < 0:
                raise ValueError(f"The parameter kappa needs to be greater or equal to 0. Supplied kappa is {value}.")
            self._kappa = value

    def _setup_parameters(self) -> None:
        """

        :return:
        """
        n_x = self._model.n_x

        self._lambda = self._alpha ** 2 * (n_x + self._kappa) - n_x
        self._gamma = np.sqrt(n_x + self._lambda)

        weights = np.zeros((2, 2 * n_x + 1))
        weights[0, 0] = self._lambda / (n_x + self._lambda)
        weights[1, 0] = self._lambda / (n_x + self._lambda) + 1 - self._alpha ** 2 + self._beta
        weights[:, 1:] = 1 / (2 * (n_x + self._lambda))
        self._weights = weights

        a = ca.SX.sym('a', (n_x, n_x))
        self._sqrt = ca.Function('sqrt', [a], [ca.chol(a)])

    def _setup_predict(self, *args: Optional[ca.Function]) -> None:
        """

        :param args:
        :return:
        """
        n_x = self._model.n_x
        n_u = self._model.n_u
        n_p = self._model.n_p

        x = ca.MX.sym('x', n_x)
        u = ca.MX.sym('u', n_u)
        p = ca.MX.sym('p', n_p)
        up = ca.vertcat(u, p)

        P = ca.MX.sym('P', (n_x, n_x))
        Q = ca.MX.sym('Q', (n_x, n_x))
        P_sqrt = self._sqrt(P)
        X = [x]
        for k in range(n_x):
            X.append(x + self._gamma * P_sqrt[:, k])
        for k in range(n_x):
            X.append(x - self._gamma * P_sqrt[:, k])
        X = ca.horzcat(*X)

        self._solution.setup('P', P={
            'values_or_names': [P.name() + '_' + str(k) for k in range(P.numel())],
            'description': P.numel() * [''],
            'labels': P.numel() * [''],
            'units': P.numel() * [''],
            'shape': (P.numel(), 0),
            'data_format': ca.DM
        })

        sol = self._model(x0=X, p=up)
        X = sol['xf']

        x_pred = ca.MX(0)
        for k in range(2 * n_x + 1):
            x_pred += self._weights[0, k] * X[:, k]

        P_pred = Q
        for k in range(2 * n_x + 1):
            P_pred += self._weights[1, k] * (X[:, k] - x_pred) @ (X[:, k] - x_pred).T

        self._predict_function = ca.Function('prediction_step',
                                             [ca.horzcat(x, P), up, Q],
                                             [ca.horzcat(x_pred, P_pred, X)],
                                             ['x0', 'p', 'Q'],
                                             ['x'])

    def _setup_update(self, *args: Optional[ca.Function]) -> None:
        """

        :param args:
        :return:
        """
        n_x = self._model.n_x
        n_y = self._model.n_y
        n_u = self._model.n_u
        n_p = self._model.n_p

        x = ca.MX.sym('x', n_x)
        X = ca.MX.sym('X', n_x, 2 * n_x + 1)
        y = ca.MX.sym('y', n_y)
        u = ca.MX.sym('u', n_u)
        p = ca.MX.sym('p', n_p)
        up = ca.vertcat(u, p)
        P = ca.MX.sym('P', (n_x, n_x))

        if n_y == 0:
            warnings.warn(f"The model has no measurement equations, I am assuming measurements of all states "
                          f"{self._model.dynamical_state_names} are available.")
            Y = X
        else:
            sol = self._model(x0=X, p=up, which='meas_function')
            Y = sol['yf']

        y_pred = ca.MX(0)
        for k in range(2 * n_x + 1):
            y_pred += self._weights[0, k] * Y[:, k]

        R = ca.MX.sym('R', (n_y, n_y))
        P_xy = ca.MX(0)
        P_yy = R
        for k in range(2 * n_x + 1):
            P_xy += self._weights[1, k] * (X[:, k] - x) @ (Y[:, k] - y_pred).T
            P_yy += self._weights[1, k] * (Y[:, k] - y_pred) @ (Y[:, k] - y_pred).T

        # NOTE: See NOTE in KalmanFilter.setup()
        K = ca.solve(P_yy.T, P_xy.T).T

        x_up = x + K @ (y - y_pred)
        P_up = P - K @ P_yy @ K.T

        self._update_function = ca.Function('update_step',
                                            [ca.horzcat(x, P, X), y, up, R],
                                            [ca.horzcat(x_up, P_up), y_pred],
                                            ['x0', 'y', 'p', 'R'],
                                            ['x', 'y_pred'])

    @property
    def alpha(self) -> float:
        """

        :return:
        """
        return float(self._alpha)

    @alpha.setter
    def alpha(self, arg: Union[int, float]) -> None:
        self._check_parameter_bounds('alpha', arg)

    @property
    def beta(self) -> float:
        """

        :return:
        """
        return float(self._beta)

    @beta.setter
    def beta(self, arg: Union[int, float]) -> None:
        self._check_parameter_bounds('beta', arg)

    @property
    def kappa(self) -> float:
        """

        :return:
        """
        return float(self._kappa)

    @kappa.setter
    def kappa(self, arg: Union[int, float]) -> None:
        self._check_parameter_bounds('kappa', arg)


__all__ = [
    'KalmanFilter',
    'ExtendedKalmanFilter',
    'UnscentedKalmanFilter'
]



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from importlib import import_module, invalidate_caches
import os
import shutil
from subprocess import call
import sys
from types import ModuleType

import casadi as ca
import numpy as np

from muaompc import ldt


mixed_box_constraints_reg_orig = """variable u[0:N-1](m);
auxs x[0:N](n);
parameters x0(n);
minimize sum(quad(x[i],Q)+quad(u[i], R), i=0:N-1)+quad(x[N],P);
subject to x[i+1] = A*x[i]+B*u[i], i=0:N-1;
u_lb <= u[i] <= u_ub, i=0:N-1;
x_lb <= x[i] <= x_ub, i=0:N-1;
x[0]=x0;
"""


CTL_SOLVER = None
HILO_MPC = None


def solver(x0):
    """

    :param x0:
    :return:
    """
    global CTL_SOLVER, HILO_MPC
    x0_shape = CTL_SOLVER.parameters.x0.shape
    n_u = HILO_MPC._n_u

    CTL_SOLVER.parameters.x0[:] = np.array(x0).reshape(x0_shape)
    CTL_SOLVER.solve_problem()
    return CTL_SOLVER.u_opt[0:n_u]


def set_ctl_solver(ctl):
    """

    :param ctl:
    :return:
    """
    global CTL_SOLVER
    CTL_SOLVER = ctl


def set_hilo_mpc(mpc):
    """

    :param mpc:
    :return:
    """
    global HILO_MPC
    HILO_MPC = mpc


def setup_solver(mpc):
    """

    :param mpc:
    :return:
    """
    dest_dir = 'muaompc_codegen'
    mod_name = 'hiloext'
    prb_name = 'muaompc'
    os.makedirs(dest_dir, exist_ok=True)
    prb_fname = os.path.join(dest_dir, prb_name+'.prb')
    basedir = shutil.os.path.abspath(shutil.os.curdir)
    codegendir = os.path.join(basedir, dest_dir, '%s_%s' % (
        mod_name, prb_name))
    is_new_prb = write_prb_file(prb_fname,
                                mixed_box_constraints_reg_orig)
    datmod = gen_data_module(mpc)

    # Generate first code for microcontroller
    mpcprb = ldt.setup_mpc_problem(
        prb_fname, prefix=mod_name, destdir=dest_dir,
        numeric='float32')  # single precision float for microcontrollers
    ldt.generate_mpc_data(mpcprb, datmod, safe_mode=False, muc=True)

    # Generate Cython interface
    mpcprb = ldt.setup_mpc_problem(
        prb_fname, prefix=mod_name, destdir=dest_dir,
        numeric='float64')  # Cython interface uses double precision float
    if is_new_prb:
        compile_controller(mod_name, basedir, codegendir)
        m = "I must stop the execution of the program. Restart your program "
        m += " to correctly load the generated muaompc interface module. "
        m += "muaompc's Python interface was installed successfully.\n"
        m += "After restarting your program, you should not see this message."
        # It is not trivial to _universally_ load a module installed on the fly
        raise ModuleNotFoundError(m)
    ldt.generate_mpc_data(mpcprb, datmod, safe_mode=False, muc=False)
    ctl = init_controller(datmod, mod_name, basedir, codegendir)
    config_controller(ctl, mpc._nlp_opts)
    set_hilo_mpc(mpc)
    set_ctl_solver(ctl)
    return solver


def compile_controller(mod_name, basedir, codegendir):
    """

    :param mod_name:
    :param basedir:
    :param codegendir:
    :return:
    """
    python = sys.executable
    shutil.os.chdir(codegendir)
    call([python, mod_name + 'setup.py', 'install'])
    invalidate_caches()
    shutil.os.chdir(basedir)


def config_controller(ctl, opts):
    """

    :param ctl:
    :param opts:
    :return:
    """
    if opts.get('in_iter') is not None:
        ctl.conf.in_iter = opts['in_iter']
    if opts.get('ex_iter') is not None:
        ctl.conf.ex_iter = opts['ex_iter']
    if opts.get('warm_start') is not None:
        ctl.conf.warm_start = opts['warm_start']


def init_controller(dat_mod, mod_name, basedir, codegendir):
    """

    :param dat_mod:
    :param mod_name:
    :param basedir:
    :param codegendir:
    :return:
    """
    ctlmod = import_module('.'+mod_name+'ctl', package=mod_name)
    data_name = dat_mod.__name__
    shutil.os.chdir(codegendir)
    ctl = ctlmod.Ctl('data/%s/%s%s.json' % (data_name, mod_name, data_name))
    shutil.os.chdir(basedir)
    return ctl


def write_prb_file(fname, prb):
    """

    :param fname:
    :param prb:
    :return:
    """
    try:
        f = open(fname, 'r')
    except FileNotFoundError:
        write_new_prb = True
    else:
        if f.read() == prb:  # content of file is identical
            write_new_prb = False  # do not overwrite prb file
        else:
            write_new_prb = True
        f.close()

    if write_new_prb:
        with open(fname, 'w') as f:
            f.write(prb)

    return write_new_prb


def gen_data_module(mpc):
    """

    :param mpc:
    :return:
    """
    model = mpc._model
    #R = np.array(mpc.R).reshape(m,m)
    dat = ModuleType('dat')
    dat.A = np.array(ca.DM(model.A))
    dat.B = np.array(ca.DM(model.B))
    dat.Q = np.array(ca.DM(mpc.Q))
    try:
        dat.P = np.array(ca.DM(mpc.P))
    except AttributeError:
        dat.P = np.array(ca.DM(mpc.Q))
    dat.R = np.array(ca.DM(mpc.R))
    dat.n, dat.m = dat.B.shape
    dat.u_lb = np.array(ca.DM(mpc._u_lb)).reshape(dat.m, 1)
    dat.u_ub = np.array(ca.DM(mpc._u_ub)).reshape(dat.m, 1)
    dat.x_lb = np.array(ca.DM(mpc._x_lb)).reshape(dat.n, 1)
    dat.x_ub = np.array(ca.DM(mpc._x_ub)).reshape(dat.n, 1)
    dat.N = mpc.horizon

    return dat



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from abc import ABCMeta, abstractmethod


class Controller(metaclass=ABCMeta):
    """"""
    _type = ''

    def __init__(self, *args, **kwargs):
        """Constructor method"""
        super().__init__(*args, **kwargs)
        self._update_type()

    @abstractmethod
    def _update_type(self) -> None:
        """

        :return:
        """
        pass

    @property
    def type(self) -> str:
        """

        :return:
        """
        return self._type



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

from typing import Optional, TypeVar, Union

import casadi as ca
import numpy as np
from scipy.optimize import minimize


Numeric = Union[int, float]
Symbolic = TypeVar('Symbolic', ca.SX, ca.MX)


class IpoptDebugger(ca.Callback):
    """"""
    def __init__(self, name, n_v, n_g, n_p, iter_step, ind_x, ind_u, opts=None):
        """Constructor method"""
        super().__init__()

        self.n_v = n_v
        self.n_g = n_g
        self.np = n_p
        self.ind_x = ind_x
        self.ind_u = ind_u
        self.x_sols = []
        self.f_sols = []
        self.g_sols = []
        self.lam_x_sols = []
        self.lam_g_sols = []
        self.lam_p_sols = []
        self.iter = []
        self.iter_step = iter_step

        # Initialize internal objects
        if opts is None:
            opts = {}
        self.construct(name, opts)

    def get_n_in(self):
        """

        :return:
        """
        return ca.nlpsol_n_out()

    def get_n_out(self):
        """

        :return:
        """
        return 1

    def get_name_in(self, i):
        """

        :param i:
        :return:
        """
        return ca.nlpsol_out(i)

    def get_name_out(self, i):
        """

        :param i:
        :return:
        """
        return "ret"

    def get_sparsity_in(self, i):
        """

        :param i:
        :return:
        """
        n = ca.nlpsol_out(i)
        if n == 'f':
            return ca.Sparsity.scalar()
        elif n in ('x', 'lam_x'):
            return ca.Sparsity.dense(self.n_v)
        elif n in ('g', 'lam_g'):
            return ca.Sparsity.dense(self.n_g)
        else:
            return ca.Sparsity(0, 0)

    def eval(self, arg):
        """

        :param arg:
        :return:
        """
        # Create dictionary
        if not self.iter:
            self.iter = [0]
        else:
            self.iter.append(self.iter[-1] + self.iter_step)
        darg = {}
        for (i, s) in enumerate(ca.nlpsol_out()): darg[s] = arg[i]
        x_sol = [float(elem) for elem in darg['x'].full()]

        self.x_sols.append(x_sol)
        self.f_sols.append(darg['f'].full()[0][0])
        self.g_sols.append(darg['g'].full())
        self.lam_x_sols.append(darg['lam_x'].full())
        self.lam_g_sols.append(darg['lam_g'].full())
        self.lam_p_sols.append(darg['lam_p'].full())

        return [0]

    def reset_solution(self):
        """

        :return:
        """
        self.x_sols = []
        self.f_sols = []
        self.g_sols = []
        self.lam_x_sols = []
        self.lam_g_sols = []
        self.lam_p_sols = []


class SciPyOptimizer:
    """
    Wrapper class for SciPy's optimizations

    :param name:
    :type name: str
    :param solver:
    :type solver: str
    :param problem:
    :type problem:
    :param options:
    :type options:
    """
    def __init__(
            self,
            name: str,
            solver: str,
            problem: dict[str, Symbolic],
            options: Optional[dict[str, Union[str, Numeric]]] = None
    ) -> None:
        """Constructor method"""
        self._name = name
        self._solver = solver

        self._function = None
        self._jacobian = None
        self._hessian = None

        self._get_free = []
        self._has_free = False

        self._n_x = 0
        self._n_p = 0
        self._n_f = 0

        self._parse_problem(problem)

        if options is None:
            options = {}
        self._options = options

        self._stats = {}

    def __call__(self, *args, **kwargs):
        """Calling method"""
        # TODO: Dimension checks
        # TODO: Typing hints
        if args and kwargs:
            raise TypeError("Instance of SciPyOptimizer can only accept positional or keyword arguments, not both")

        if args:
            if len(args) == 1:
                x0 = args[0]
                p = None
                lbx = None
                ubx = None
            elif len(args) == 2:
                x0 = args[0]
                p = args[1]
                lbx = None
                ubx = None
            elif len(args) == 3:
                x0 = args[0]
                p = args[1]
                lbx = args[2]
                ubx = None
            elif len(args) == 4:
                x0 = args[0]
                p = args[1]
                lbx = args[2]
                ubx = args[3]
            else:
                raise ValueError()

        if kwargs:
            x0 = kwargs.get('x0')
            p = kwargs.get('p')
            lbx = kwargs.get('lbx')
            ubx = kwargs.get('ubx')

        if not args and not kwargs:
            x0 = np.zeros((self._n_x, 1))
            if self._n_p > 0:
                p = np.zeros((self._n_p, 1))
            else:
                p = None

        min_kwargs = {}
        if p is not None:
            min_kwargs['args'] = (p, )
        min_kwargs['method'] = self._solver
        if self._solver in ['CG', 'BFGS', 'Newton-CG', 'L-BFGS-B']:
            min_kwargs['jac'] = self._jacobian
        if self._solver in ['Newton-CG']:
            min_kwargs['hess'] = self._hessian
        if self._solver in ['L-BFGS-B']:
            bounds = []
            for k in range(self._n_x):
                lbk = lbx[k] if lbx is not None else None
                ubk = ubx[k] if ubx is not None else None
                bounds.append((lbk, ubk))
            min_kwargs['bounds'] = bounds
        min_kwargs.update(self._options)

        # Ensure x0 is 1D for scipy.optimize.minimize (required in scipy 1.11+)
        x0_flat = np.asarray(x0).flatten()
        sol = minimize(self._function, x0_flat, **min_kwargs)

        self._stats['status'] = sol.status
        self._stats['success'] = sol.success
        self._stats['message'] = sol.message

        return {'f': sol.fun, 'x': sol.x}

    def _parse_problem(self, problem: dict[str, Symbolic]) -> None:
        """

        :param problem:
        :return:
        """
        # TODO: Add support for constraints
        # TODO: Add support for multi-objective optimization
        f = problem.get('f')
        x = problem.get('x')
        p = problem.get('p')

        if f is None:
            raise ValueError("No objective function supplied")
        if x is None:
            raise ValueError("No decision variables supplied")
        if p is None:
            p = f.sym('', 0)

        self._n_x = x.shape[0]
        self._n_p = p.shape[0]
        self._n_f = f.shape[0]

        function = ca.Function('function', [x, p], [f])
        function_free = function.get_free()

        def fun(w, args):
            """

            :param w:
            :param args:
            :return:
            """
            return function(w, args).full().flatten()

        self._function = fun

        if self._solver in ['CG', 'BFGS', 'Newton-CG', 'L-BFGS-B']:
            jacobian = ca.Function('jacobian', [x, p], [ca.jacobian(f, x)])
            jacobian_free = jacobian.get_free()

            def jac(w, args):
                """

                :param w:
                :param args:
                :return:
                """
                return jacobian(w, args).full().flatten()

            self._jacobian = jac
        else:
            jacobian_free = []

        if self._solver in ['Newton-CG']:
            hessian = ca.Function('hessian', [x, p], [ca.hessian(f, x)[0]])
            hessian_free = hessian.get_free()

            def hess(w, args):
                """

                :param w:
                :param args:
                :return:
                """
                return hessian(w, args).full()

            self._hessian = hess
        else:
            hessian_free = []

        self._get_free = list(set(function_free + jacobian_free + hessian_free))
        if self._get_free:
            self._has_free = True

    def get_free(self) -> list[str]:
        """

        :return:
        """
        return self._get_free

    def has_free(self) -> bool:
        """

        :return:
        """
        return self._has_free

    def stats(self) -> dict[str, Union[bool, int, str]]:
        """

        :return:
        """
        return self._stats



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from typing import Optional, Sequence, Union

import casadi as ca
import numpy as np

from .base import Controller
from ..base import Base, TimeSeries
from ...util.util import convert, is_diagonal


Numeric = Union[int, float]
Array = Union[Sequence[Numeric], np.ndarray]


class PID(Controller, Base):
    """
    Class for PID controller

    :param n_set_points:
    :type n_set_points: int
    :param id:
    :type id: str, optional
    :param name:
    :type name: str, optional
    :param k_p:
    :type k_p:
    :param t_i:
    :type t_i:
    :param t_d:
    :type t_d:
    :param proportional_on_process_value:
    :type proportional_on_process_value: bool
    :param derivative_on_process_value:
    :type derivative_on_process_value: bool
    :param plot_backend:
    :type plot_backend: str, optional
    """
    def __init__(
            self,
            n_set_points: int = 1,
            id: Optional[str] = None,
            name: Optional[str] = None,
            k_p: Optional[Union[Numeric, Array]] = None,
            t_i: Optional[Union[Numeric, Array]] = None,
            t_d: Optional[Union[Numeric, Array]] = None,
            proportional_on_process_value: bool = False,
            derivative_on_process_value: bool = False,
            plot_backend: Optional[str] = None
    ) -> None:
        """Constructor method"""
        super().__init__(id=id, name=name)
        if self._id is None:
            self._create_id()

        self._n_set_points = n_set_points
        if k_p is None:
            k_p = self._n_set_points * [1.]
        self._set_tuning_parameter('k_p', k_p)
        if t_i is None:
            t_i = self._n_set_points * [np.inf]
        self._set_tuning_parameter('t_i', t_i)
        if t_d is None:
            t_d = self._n_set_points * [0.]
        self._set_tuning_parameter('t_d', t_d)
        self._proportional_on_process_value = proportional_on_process_value
        self._derivative_on_process_value = derivative_on_process_value
        self._p_band = False
        self._anti_windup = None
        self._set_point = ca.DM.zeros(self._n_set_points)

        self._solution = TimeSeries(plot_backend, parent=self)

    def _update_type(self) -> None:
        """

        :return:
        """
        self._type = 'PID'

    def _set_tuning_parameter(self, which: str, value: Union[Numeric, Array]) -> None:
        """

        :param which:
        :param value:
        :return:
        """
        # TODO: Dimensions
        value = convert(value, np.ndarray, shape=(self._n_set_points, self._n_set_points))
        if not is_diagonal(value):
            raise ValueError(f"The number of set points is greater than 1, but the supplied matrix for "
                             f"{dict({'k_p': 'K_P', 't_i': 'T_I', 't_d': 'T_D'})[which]} is not a diagonal matrix. "
                             f"Coupled multi-variable control is not supported at the moment.")
        if which == 'k_p':
            self._k_p = np.diag(value)
        elif which == 't_i':
            self._t_i = np.diag(value)
        elif which == 't_d':
            self._t_d = np.diag(value)

    def _append_process_value(self, process_value: ca.DM) -> None:
        """

        :param process_value:
        :return:
        """
        if self._solution.get_by_id('x').is_empty():
            process_value_vector = ca.DM.zeros(self._n_set_points, 3)
        else:
            process_value_vector = ca.reshape(self._solution.get_by_id('x:f'), self._n_set_points, 3)
            process_value_vector[:, :-1] = process_value_vector[:, 1:]
        process_value_vector[:, -1] = process_value
        self._solution.add('x', process_value_vector.reshape((3 * self._n_set_points, -1)))

    def _append_set_point(self) -> None:
        """

        :return:
        """
        if self._solution.get_by_id('p').is_empty():
            set_point_vector = ca.DM.zeros(self._n_set_points, 3)
        else:
            params = self._solution.get_by_id('p:f')
            set_point_vector = ca.reshape(params[:3 * self._n_set_points], self._n_set_points, 3)
            set_point_vector[:, :-1] = set_point_vector[:, 1:]
        set_point_vector[:, -1] = self._set_point
        params = ca.vertcat(set_point_vector.reshape((3 * self._n_set_points, -1)), self._k_p, self._t_i, self._t_d)
        self._solution.add('p', params)

    def _initialize_controller_output(self) -> None:
        """

        :return:
        """
        u = ca.DM.zeros(self._n_set_points)
        self._solution.set('u', u)

    @property
    def proportional_gain(self) -> np.ndarray:
        """

        :return:
        """
        return np.diag(self._k_p)

    @proportional_gain.setter
    def proportional_gain(self, k_p: Union[Numeric, np.ndarray]) -> None:
        self._set_tuning_parameter('k_p', k_p)

    k_p = proportional_gain

    @property
    def integral_time(self) -> np.ndarray:
        """

        :return:
        """
        return np.diag(self._t_i)

    @integral_time.setter
    def integral_time(self, t_i: Union[Numeric, np.ndarray]) -> None:
        self._set_tuning_parameter('t_i', t_i)

    t_i = integral_time

    @property
    def integral_gain(self) -> np.ndarray:
        """

        :return:
        """
        return np.diag(self._k_p / self._t_i)

    k_i = integral_gain

    @property
    def derivative_time(self) -> np.ndarray:
        """

        :return:
        """
        return np.diag(self._t_d)

    @derivative_time.setter
    def derivative_time(self, t_d: Union[Numeric, np.ndarray]) -> None:
        self._set_tuning_parameter('t_d', t_d)

    t_d = derivative_time

    @property
    def derivative_gain(self) -> np.ndarray:
        """

        :return:
        """
        return np.diag(self._k_p * self._t_d)

    k_d = derivative_gain

    @property
    def set_point(self) -> ca.DM:
        """

        :return:
        """
        return self._set_point

    @set_point.setter
    def set_point(self, set_point: Union[Numeric, np.ndarray]) -> None:
        set_point = convert(set_point, ca.DM)
        if self._n_set_points > 1:
            if set_point.size1() == 1:
                set_point = ca.repmat(set_point, self._n_set_points, 1)
            elif set_point.size1() != self._n_set_points:
                raise ValueError(f"Dimension mismatch. Supplied dimension for the set point is "
                                 f"{set_point.size1()}x{set_point.size2()}, but required dimension is "
                                 f"{self._n_set_points}x1.")
        self._set_point = set_point

    s_p = set_point

    def setup(self, **kwargs) -> None:
        """

        :param kwargs:
        :return:
        """
        dt = kwargs.get('dt')
        if dt is None:
            dt = 1.

        process_value = ca.SX.sym('pv', self._n_set_points, 3)
        set_point = ca.SX.sym('sp', self._n_set_points, 3)
        error = set_point - process_value
        output = ca.SX.sym('u', self._n_set_points)
        delta_output = 0.

        # TODO: Dimensions
        k_p = ca.SX.sym('k_p', self._n_set_points)
        t_i = ca.SX.sym('t_i', self._n_set_points)
        t_d = ca.SX.sym('t_d', self._n_set_points)

        k_p_matrix = ca.diag(k_p)
        t_i_matrix_inv = ca.diag(1. / t_i)
        t_d_matrix = ca.diag(t_d)

        if self._p_band:
            raise NotImplementedError("The use of the proportional band has not been implemented yet.")
        else:
            if self._proportional_on_process_value:
                delta_output -= process_value[:, -1] - process_value[:, -2]
            else:
                delta_output += error[:, -1] - error[:, -2]

        delta_output += dt * t_i_matrix_inv @ error[:, -1]
        # TODO: Add anti-windup

        if self._derivative_on_process_value:
            delta_output -= (t_d_matrix / dt) @ (process_value[:, -1] - 2 * process_value[:, -2] + process_value[:, -3])
        else:
            delta_output += (t_d_matrix / dt) @ (error[:, -1] - 2 * error[:, -2] + error[:, -3])

        # TODO: Add anti-windup
        parameters = ca.vertcat(set_point.reshape((3 * self._n_set_points, -1)), k_p, t_i, t_d)
        self._function = ca.Function(
            'function',
            [process_value.reshape((3 * self._n_set_points, -1)), ca.vertcat(output, parameters)],
            [error, output + k_p_matrix @ delta_output],
            ['x0', 'p'],
            ['error', 'yf']
        )

        self.check_consistency()

        n_pv = process_value.numel()
        n_y = output.numel()
        n_sp = set_point.numel()
        n_k_p = k_p.numel()
        n_t_i = t_i.numel()
        n_t_d = t_d.numel()
        n_p = n_sp + n_k_p + n_t_i + n_t_d  # same as parameters.numel()

        names = ['dt', 'x', 'u', 'p']  # 'y'
        vector = {
            'dt': dt,
            'x': {
                'values_or_names': [pv.name() for pv in process_value.elements()],
                'description': n_pv * [''],
                'labels': n_pv * ['process value'],
                'units': n_pv * [''],
                'shape': (n_pv, 0),
                'data_format': ca.DM
            },
            'u': {
                'values_or_names': [out.name() for out in output.elements()],
                'description': n_y * [''],
                'labels': n_y * ['controller output'],
                'units': n_y * [''],
                'shape': (n_y, 0),
                'data_format': ca.DM
            },
            'p': {
                'values_or_names': [par.name() for par in parameters.elements()],
                'description': n_p * [''],
                'labels': n_sp * ['set point'] + n_k_p * ['K_P'] + n_t_i * ['T_I'] + n_t_d * ['T_D'],
                'units': n_p * [''],
                'shape': (n_p, 0),
                'data_format': ca.DM
            }
        }
        self._solution.setup(*names, **vector)

    def call(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        if self._function is None:
            type_ = 'P'
            if self._t_i != np.inf:
                type_ += 'I'
            if self._t_d != 0.:
                type_ += 'D'
            raise RuntimeError(f"{type_} controller is not set up. Run PID.setup() before calling the {type_} "
                               f"controller")

        pv = kwargs.get('pv')
        if pv is not None:
            pv = convert(pv, ca.DM)
        else:
            pv = ca.DM.zeros(self._n_set_points, 1)
        self._append_process_value(pv)
        self._append_set_point()
        if self._solution.get_by_id('u').is_empty():
            self._initialize_controller_output()

        args = self._solution.get_function_args()
        result = self._function(**args)
        self._solution.update(u=result['yf'])

        return result['yf']


__all__ = [
    'PID'
]
